def main():
    print("Hello, it's your task assistant Cellai")
    print("I'm still in development, but my creators are happy to share updates.")
    print("Feel free to reach out to them at <welcome@neuronum.net> for more details.")

if __name__ == "__main__":
    main()
